# -*- coding: utf-8 -*-

########################################################################
#
# License: BSD
# Created: October 19, 2004
# Author:  Ivan Vilata i Balaguer - reverse:net.selidor@ivan
#
# $Id$
#
########################################################################

"""Special node behaviours for PyTables.

This package contains several modules that give specific behaviours
to PyTables nodes.  For instance, the filenode module provides
a file interface to a PyTables node.


Package modules:
  filenode -- A file interface to nodes for PyTables databases.

"""

# The list of names to be exported to the importing module.
__all__ = ['filenode']


## Local Variables:
## mode: python
## py-indent-offset: 4
## tab-width: 4
## End:
